from .encoders_map import encoders_map
from .networks_map import networks_map
from .optimizers_maps import optimizers_map

from .acceptable_frames_map import acceptable_frames_map

from .encoders_map import encoders_map as encoder_map
from .networks_map import networks_map as network_map
from .optimizers_maps import optimizers_map as optimizer_map
