from .Aftab import Aftab

from .constants import seeds
from .constants import seeds as SEEDS
from .constants import seeds as aftab_seeds

from .constants import environments
from .constants import environments as ENVS
from .constants import environments as aftab_environments

__version__ = "0.1.58"
